(defvar *homebases* nil)
(defvar *known-homebase-count* 0)
(defvar *known-homebase-weights* 10)

(defstruct homebase
  (x 0 :type fixnum)
  (y 0 :type fixnum))

(defun sum-homebase-packages (homebase)
  (let* ((x (homebase-x homebase))
         (y (homebase-y homebase))
         (pkgs (packages-at x y))
         (weights (mapcar #'game-package-weight pkgs))
         (ok-weights (remove-if-not #'identity weights)))
    (reduce #'+ ok-weights :initial-value 0)))

(defun homebase-weight-estimate (homebase)
  (declare (ignore homebase))
  ;; Evil floatification operation
  ;; A small epsilon always makes exploring unknown bases
  ;; worthwhile
  (+ 0.001
     (/ (* *known-homebase-weights* 1.0)
        (if (zerop *known-homebase-count*) 1
            *known-homebase-count*))))

(defun maybe-update-homebase-stats (x y)
  (declare (fixnum x y))
  (let ((base (find-if (lambda (base)
                         (and (= (homebase-x base) x)
                              (= (homebase-y base) y)))
                       *homebases*)))
    (when base
      ;; Update the stats
      (when (zerop *known-homebase-count*)
        (setf *known-homebase-weights* 0))
      (incf *known-homebase-count*)
      (incf *known-homebase-weights*
            (sum-homebase-packages base)))

    ;; Once we have obtained all of the possible package information
    ;; available at the homebase, we can ditch the homebase info
    (setf *homebases* (delete base *homebases*))))

(defun clean-unreachable-homebases ()
    (setf *homebases*
          (delete-if-not (lambda (base)
                           (reachable (homebase-x base) (homebase-y base)))
                         *homebases*)))

(defun maybe-sort-and-snip-bases (bases x y)
  (flet ((sort-fn (base)
           (manhattan-heuristic x y
                                (homebase-x base)
                                (homebase-y base))))
    (if (<= (length bases) 100) bases
        (subseq (sort (copy-list bases) #'< :key #'sort-fn) 0 500))))