;;; These few functions are for interactive testing
;;; and debugging only
;;; They have no official purpose
(defun setup-env ()
  (push '(:lisp_icfp_host . "icfp1.cse.ogi.edu") *environment-list*)
  (push '(:lisp_icfp_port . "22101") *environment-list*))

(defun clear-state ()
  (setf *homebases* nil)
  (setf *known-homebase-count* 0)
  (setf *known-homebase-weights* 10)

  (setf *packages* (make-hash-table :test #'eq))
  (setf *known-package-count* 0)
  (setf *known-package-weights* 1)
  (setf *packages-at* (make-hash-table :test #'equal))

  (setf *robots* (make-hash-table :test #'eq))
  (setf *live-robots* nil)

  (setf *turn* 0))

(defvar *netcat-process* nil)
(defun kill-netcat ()
  (when *netcat-process*
    (process-kill *netcat-process* :sigterm)
    (setf *netcat-process* nil)))

;;; THE MAIN PROGRAM
(defun main ()
  (let* ((host (cdr (assoc :LISP_ICFP_HOST ext:*environment-list*)))
         (port (cdr (assoc :LISP_ICFP_PORT ext:*environment-list*)))
         ;; Start nc
         (process (ext:run-program "source/nc/nc" (list host port)
                                   :wait nil
                                   :input :stream
                                   :output :stream)))

    ;;; Give netcat enough time to start
    (sleep 1)

    (setf *netcat-process* process)

    (unless process
      (write-line "Could not start netcat, bailing out" *debug-io*)
      (ext:quit))

    (unless (process-alive-p process)
      (write-line "Netcat died early, bailing out" *debug-io*)
      (ext:quit))

    ;;; OK, it's time to get dirty
    (ignore-errors
      (setf (process-status-hook process) #'handle-exit)
      (let ((*standard-input* (process-output process))
            (*standard-output* (process-input process))
            ;; We try to avoid gc:ing too ofter
            ;; Also, we _try_ to save space, but
            ;; A* for example leads us straight to
            ;; consing heaven
            ;; 10 MB should be enough.
            (*bytes-consed-between-gcs* (* 10 1024 1024)))

        (initialize-world-and-configuration)
        ;; Update reachability map
        (setf *reachable* (make-reachability-table (our-bot-x) (our-bot-y)))

        (clean-unreachable-homebases)
        (play-the-game)))
    (exit-the-system)))

(defun handle-exit (process)
  (unless (eq (process-status process) :running)
    (exit-the-system)))

(defun exit-the-system ()
  (when *netcat-process*
    (process-kill *netcat-process* :sigterm)
    (process-close *netcat-process*)
    (ext:quit)))

(defun initialize-world-and-configuration ()
  (write-line "Player")
  (force-output)
  (read-board)
  (when +debugging+
    (print-board))
  (read-configuration)
  (handle-server-response (chop-server-response (read-server-response))))

(defun bid-enemy-near ()
  ;; Offer 0.1% of fuel left or at least 15 points,
  ;; however, not more than 20% of available fuel
  (min
   (max (round (* 0.001 *capital*)) 15)
   (round (* 0.2 *capital*))))

(defun bid-enemy-pushable ()
  ;; Offer 1% of the fuel or at least 100 points
  ;; however, not more than 40% of available fuel
  (min
   (max (round (* 0.01 *capital*)) 100)
   (round (* 0.4 *capital*))))

;;; The main loop
(defun play-the-game ()
  (let (smallest packages-to-here previous-op
                 route can-pick-up
                 bot-near bid
                 direction-to-push-someone-to-water
                 local-plan)
    (loop
     (tagbody
        (debug-print :x (our-bot-x)
                     :y (our-bot-y))
        ;; Bookkeeping
        (incf *turn*)
        (setf smallest (handle-package-list (read-list-of-packages)))
        (setf packages-to-here (find-packages-destined-here))
        (setf bid 1)
        (setf bot-near (robot-near-us))

        (debug-print :homebase-length (length *homebases*))

        (when bot-near
          ;; Increase the bidding so that we get to do what
          ;; we want to do
          (setf bid (bid-enemy-near)))

        ;; Strategy 0 - DROP
        (when packages-to-here
          (drop-packages packages-to-here bid)
          (setf route nil)
          (setf previous-op nil)
          (go server-response))

        ;; Strategy 1/2
        ;; Chasing others
        (when *chasing-others*
          (debug-print :chasing-others)
          (let ((x1 (our-bot-x))
                (y1 (our-bot-y))
                (x2 (robot-x *chasing-others*))
                (y2 (robot-y *chasing-others*)))
            (when (< (manhattan-heuristic x1 y1 x2 y2)
                     +lookout-window+)
              (let* ((path (find-path-localized x1 y1 x2 y2))
                     (step (and path (last (path-reversed path)))))
                (when step
                  (setf previous-op step)
                  (send-move bid previous-op)
                  (go server-response))))))

        ;; In hostile waters, part deux
        (when bot-near
          (when (we-were-pushed previous-op)
            ;; retarget and move on
            (debug-print :retargeting-due-to-push)
            (setf route nil)
            (go acquire-target))

          ;; Just check this one bot for pushing, we are not _that_
          ;; violent
          (setf direction-to-push-someone-to-water (bot-near-water bot-near))
          (when direction-to-push-someone-to-water
            (debug-print :attempting-to-push-someone)
            (setf bid (bid-enemy-pushable))
            (send-move bid direction-to-push-someone-to-water)
            (setf route nil)
            (setf previous-op nil)
            (go server-response)))

      enemy-handled
        ;; Recalculate carrying capacity
        (setf *carrying* (carrying-at-the-moment))
        (debug-print :carrying *carrying* :smallest smallest)
        (setf can-pick-up (and smallest (<= (+ *carrying* smallest)
                                            *carry-capacity*)))

        (debug-print :can-pick-up can-pick-up)

        ;; Strategy 1 - PICK
        (when can-pick-up
          (pick-packages bid)
          (setf route nil)
          (setf previous-op nil)
          (go server-response))

        ;; Strategy X -- LOOKING FOR LOCAL THINGS
        ;;
        ;; We are carrying something, there's nothing pickupable =>
        ;; consider doing a sidestep
        ;; FIXME - implement
        (debug-print :global-target *global-target*)
        (when (and (plusp *carrying*)
                   ;; Do we already have some other local plan?
                   (not local-plan)
                   ;; Do we have some global target we are
                   ;; aiming at?
                   ;; Effectively a no-op, since we have a
                   ;; *global-target* whenever we are carrying
                   ;; stuff, but best be sure
                   *global-target*
                   ;; Are we so close that sidestepping now would
                   ;; be futile?
                   (< (* 1.5 +lookout-window+)
                      (manhattan-heuristic
                       (our-bot-x)
                       (our-bot-y)
                       (first *global-target*)
                       (second *global-target*))))

          (debug-print :trying-locals)
          ;; *global-target* is updated by
          ;; by the look-for-* functions
          ;;
          ;; Check local bases first
          ;; with Manhattan, check the rest with A*
          ;; If we find a base, make a sidestep there
          ;; and set the necessary flags
          (let ((local-route (look-for-local-base)))
            (when local-route
              (debug-print :local-base local-route)
              (setf route local-route
                    local-plan t)
              (go make-a-move)))

          ;; Apparently there were no home bases.
          ;; Scan the area for packages
          ;; The pecking order here
          ;; 1) Known packages that fit and are deliverable
          ;; 2) Unknown packages, ordered by their interest rating
          ;; Look-for-local-packages will handle this ordering
          (let ((local-route (look-for-local-packages)))
            (when local-route
              (debug-print :local-package local-route)
              (setf route local-route
                    local-plan t)
              (go make-a-move))))

        ;; Strategy Y -- CHECK THAT OLD PLAN IS STILL VALID
        ;; If there has been a pick or a drop in the target area
        ;; Recompute strategy
        (unless (strategy-still-ok-p)
          (debug-print :strategy-not-ok)
          (setf route nil)
          (go acquire-target))

      acquire-target
        ;; No route - recompute
        (when (null route)
          (debug-print :retargeting)
          (setf local-plan nil)
          (setf route (find-target-global))
          (debug-print route))

        ;; Major panic - this shouldn't happen, but
        ;; make sure we don't fuck up in this case
        (when (null route)
          ;; Don't waste our ammo on a panic move
          (setf bid 1)
          (setf route (list (find-safe-map-square))))

      make-a-move
        ;; THE BASE CASE - we moved somewhere
        ;; store the direction to previous-op
        (setf previous-op (pop route))
        (send-move bid previous-op)

        ;; THE END
      server-response
        (handle-server-response
         (chop-server-response
          (read-server-response)))))))
