(defvar *packages* (make-hash-table :test 'eq
                                    :size 150))

;; Initial estimates
(defvar *known-package-count* 0)
(defvar *known-package-weights* 1)

(defmacro package (id) `(gethash ,id *packages*))
(defstruct game-package
  id
  weight
  (at-x 0 :type fixnum)
  (at-y 0 :type fixnum)
  to-x
  to-y
  undeliverable
  in-motion
  (interest-factor 1.0 :type single-float))

(defvar *packages-at* (make-hash-table :test 'equal))
(defmacro packages-at (x y) `(gethash (list ,x ,y) *packages-at*))

(defun package-weight-estimate (pkg)
  ;; Account for packages interest
  (* (game-package-interest-factor pkg)
     ;; If we know the weight, use it
     ;; Otherwise use the estimate
     (or (game-package-weight pkg)
         (/ (float *known-package-weights*)
            (if (zerop *known-package-weights*) 1
                *known-package-weights*)))))

(defun update-package-stats (weight)
  ;; If this is the first package, clear the stats
  (when (zerop *known-package-count*)
    (setf *known-package-weights* 0))

  (incf *known-package-weights* weight)
  (incf *known-package-count*))
