(defstruct robot
  id
  (x 0 :type fixnum)
  (y 0 :type fixnum)
  packages
  events)

(defmacro robot-by-id (id) `(gethash ,id *robots*))
(defvar *robots* (make-hash-table :test 'eq))

(declaim (inline our-bot-x our-bot-y))
(defun our-bot-x () (robot-x (robot-by-id *own-id*)))
(defun our-bot-y () (robot-y (robot-by-id *own-id*)))

;; Return a list of locations _next_ to a bot
(defun bot-locations ()
  (loop for bot being the hash-values of *robots*
        for x = (robot-x bot)
        for y = (robot-y bot)
        collect (list (1- x) y)
        collect (list (1+ x) y)
        collect (list x (1- y))
        collect (list x (1+ y))))

(defun robot-near-us ()
  (let ((x (our-bot-x))
        (y (our-bot-y)))
    (loop for bot being the hash-values of *robots*
          for bot-x = (robot-x bot)
          for bot-y = (robot-y bot)
          thereis (and (/= (robot-id bot) *own-id*)
                       (= (manhattan-heuristic x y bot-x bot-y) 1)
                       ;; The return value will be the first bot near
                       ;; us. There might be others, but this will
                       ;; do for most applications
                       ;; Although we might miss an oppurtunity to
                       ;; push someone else to the water
                       ;; C'est la vie.
                       bot))))

(defun bot-near-water (bot)
  (let* ((x (our-bot-x))
         (y (our-bot-y))
         (xdiff (- (robot-x bot) x))
         (ydiff (- (robot-y bot) y)))
    (when (char= (aref *world-map*
                       (+ x (* 2 xdiff))
                       (+ y (* 2 ydiff)))
                 #\~)
      (cond
        ((plusp ydiff) :n)
        ((minusp ydiff) :s)
        ((plusp xdiff) :e)
        ((minusp xdiff) :w)
        (t nil)))))
