(defvar *reachable* nil)

;; We use a flood-fill algorithm to mark the areas that are actually reachable
(defun make-reachability-table (x y)
  (let ((table (make-array (list (+ *world-width* 2)
                                 (+ *world-height* 2))
                           :element-type '(unsigned-byte 8)
                           :initial-element 0)))
    (labels
        ((groundp (i j)
           (member (aref *world-map* i j) '(#\. #\,)))
         (unmarked (i j)
           (zerop (aref table i j)))
         (maybe-mark (i j)
           (when (and (groundp i j)
                      (unmarked i j))
             (mark i j)))
         (mark (i j)
           (setf (aref table i j) 1)
           (maybe-mark (1- i) j)
           (maybe-mark (1+ i) j)
           (maybe-mark i (1- j))
           (maybe-mark i (1+ j))))
      (mark x y))
    table))

(defun reachable (x y)
    (= (aref *reachable* x y) 1))

(defun print-board ()
  (loop for y downfrom (1- (array-dimension *world-map* 1)) to 0 do
        (dotimes (x (array-dimension *world-map* 0))
          (write-char (aref *world-map* x y) *debug-io*))
        (terpri *debug-io*)))
